from .logger import RichLogger as RichLogger
from .scenarios import incremental_class as incremental_class
from .scenarios import incremental_domain as incremental_domain
from .scenarios import incremental_task as incremental_task



from typing import Any, Tuple, cast

import torch
import torch.utils.data

from avalanche.benchmarks.generators.benchmark_generators import nc_benchmark
from avalanche.benchmarks.scenarios.new_classes.nc_scenario import NCScenario


def incremental_task(
    train: torch.utils.data.Dataset[torch.Tensor],
    test: torch.utils.data.Dataset[torch.Tensor],
    transform: Any,
    n_experiences: int,
    n_classes: int,
) -> Tuple[NCScenario, int]:
    """
    Incremental task setup.

    Task labels are provided, classes from `0` to `n_classes` split over `n_experiences` (without remapping).
    Most commonly used with a multi-head setup, where each head is trained on a different experience.
    """

    benchmark = nc_benchmark(
        train_dataset=cast(Any, train),
        test_dataset=cast(Any, test),
        n_experiences=n_experiences,
        task_labels=True,
        fixed_class_order=range(n_classes),
        class_ids_from_zero_in_each_exp=True,
        train_transform=transform,
        eval_transform=transform,
    )

    n_output_classes = n_classes // n_experiences

    return benchmark, n_output_classes


def incremental_domain(
    train: torch.utils.data.Dataset[torch.Tensor],
    test: torch.utils.data.Dataset[torch.Tensor],
    transform: Any,
    n_experiences: int,
    n_classes: int,
) -> Tuple[NCScenario, int]:
    """
    Incremental domain setup.

    No task labels, classes are remapped to `0, ..., effective_n_classes`, where
    `effective_n_classes` is the number of classes per single experience.
    """

    benchmark = nc_benchmark(
        train_dataset=cast(Any, train),
        test_dataset=cast(Any, test),
        n_experiences=n_experiences,
        task_labels=False,
        fixed_class_order=range(n_classes),
        class_ids_from_zero_in_each_exp=True,
        train_transform=transform,
        eval_transform=transform,
    )

    n_output_classes = n_classes // n_experiences

    return benchmark, n_output_classes


def incremental_class(
    train: torch.utils.data.Dataset[torch.Tensor],
    test: torch.utils.data.Dataset[torch.Tensor],
    transform: Any,
    n_experiences: int,
    n_classes: int,
) -> Tuple[NCScenario, int]:
    """
    Incremental class setup.

    No task labels, classes from `0` to `n_classes` split over `n_experiences` (without remapping).
    """

    benchmark = nc_benchmark(
        train_dataset=cast(Any, train),
        test_dataset=cast(Any, test),
        n_experiences=n_experiences,
        task_labels=False,
        fixed_class_order=range(n_classes),
        train_transform=transform,
        eval_transform=transform,
    )

    n_output_classes = n_classes

    return benchmark, n_output_classes



from dataclasses import dataclass, field
from typing import Any, Dict, Iterator, List, Optional, Type, cast

import omegaconf
from hydra.conf import HydraConf, RunDir, SweepDir
from hydra.core.config_store import ConfigStore
from omegaconf import SI, DictConfig
from omegaconf.dictconfig import DictConfig
from omegaconf.omegaconf import MISSING

from .cfg import Settings


@dataclass
class Hydra(HydraConf):
    run: RunDir = RunDir("${output_dir}")
    sweep: SweepDir = SweepDir(".", "${output_dir}")
    searchpath: List[str] = field(default_factory=lambda: [
        'pkg://pytorch_yard/configs'
    ])


@dataclass
class RootConfig():
    """
    Top-level Hydra config class.
    """
    defaults: List[Any] = field(default_factory=lambda: [
        {'cfg': 'default'},
        '_self_',
        {'override hydra/job_logging': 'rich'},
        {'override hydra/hydra_logging': 'rich'},
    ])

    # Path settings
    data_dir: str = SI("${oc.env:DATA_DIR}")
    output_dir: str = SI("${oc.env:RUN_DIR}")

    # Runtime configuration
    hydra: Hydra = Hydra()

    # Experiment settings --> *.yaml
    cfg: Settings = MISSING

    # wandb metadata
    notes: Optional[str] = None
    group: Optional[str] = None
    tags: Optional[List[str]] = None


def register_configs(settings_cls: Optional[Type[Settings]] = None, settings_group: Optional[str] = None):
    """
    Register configuration options in the main ConfigStore.instance().

    The term `config` is used for a StructuredConfig at the root level (normally switchable with `-cn`
    flag in Hydra, here we use only one default config). Fields of the main config use StructuredConfigs
    with class names ending in `Settings`. `Conf` suffix is used for external schemas provided by
    the `hydra-torch` package for PyTorch/PyTorch Lightning integration, e.g. `AdamConf`.
    """
    cs = ConfigStore.instance()

    settings_group = settings_group or 'cfg'

    # Main config
    root = RootConfig()
    root.defaults[0] = {settings_group: 'default'}
    cs.store(name='root', node=DictConfig(root))

    # Config groups with defaults, YAML files validated by Python structured configs
    # e.g.: `python train.py experiment=default`
    cs.store(group=settings_group, name='settings_schema', node=settings_cls if settings_cls is not None else Settings)


def _get_tags(cfg: dict[str, Any]) -> Iterator[str]:
    for key, value in cfg.items():
        if isinstance(value, dict):
            yield from _get_tags(cast(Dict[str, Any], value))
        if key == '_tags_':
            if isinstance(value, list):
                for v in cast(List[str], value):
                    yield v
            else:
                if value is not None:
                    value = cast(str, value)
                    yield value


def get_tags(cfg: DictConfig):
    """
    Extract all tags from a nested DictConfig object.
    """
    cfg_dict = cast(Dict[str, Any], omegaconf.OmegaConf.to_container(cfg, resolve=True))
    if 'tags' in cfg_dict:
        cfg_dict['_tags_'] = cfg_dict['tags']
    return list(_get_tags(cfg_dict))



import torch

from ..core import Experiment


class PyTorchExperiment(Experiment):
    def before_entry(self) -> None:
        self.device: torch.device

    def main(self) -> None:
        pass

    def finish(self) -> None:
        pass



import os
from pathlib import Path
from typing import Any, cast

import torch
import torch.nn as nn
import torch.utils.data
from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig
from pytorch_yard.configs.cfg import Settings
from torch import Tensor

from avalanche.benchmarks.scenarios.new_classes.nc_scenario import NCScenario

from ..avalanche.wandb import WandbEpochLogger
from ..configs import get_tags
from ..core import Experiment


class AvalancheExperiment(Experiment):
    def before_entry(self) -> None:
        self.cfg: Settings
        """ Experiment config. """

        self.device: torch.device
        """ Device selector (CPU/CUDA). """

        self.train: torch.utils.data.Dataset[Tensor]
        """ Train dataset. """

        self.test: torch.utils.data.Dataset[Tensor]
        """ Test dataset. """

        self.transforms: Any
        """ Transforms applied to train and test data. """

        self.n_classes: int
        """ Number of classes in the dataset. """

        self.scenario: NCScenario
        """ Main scenario object. """

        self.model: nn.Module
        """ Main PyTorch model. """

        self.wandb_logger: WandbEpochLogger
        """ W&B logger object. """

    def main(self) -> None:
        self.wandb_logger = WandbEpochLogger(
            project=os.getenv('WANDB_PROJECT'),
            entity=os.getenv('WANDB_ENTITY'),
            name=os.getenv('RUN_NAME'),
            dir=str(Path(os.getenv('RUN_DIR', '.'))),
            config=OmegaConf.to_container(self.root_cfg, resolve=True),  # type: ignore
            tags=get_tags(cast(DictConfig, self.root_cfg)),
            notes=str(self.root_cfg.notes) if self.root_cfg.notes is not None else None,
            group=str(self.root_cfg.group) if self.root_cfg.group is not None else None,
        )

        if torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")

    def finish(self) -> None:
        self.wandb_logger.wandb.finish()  # type: ignore



__version__ = "2021.12.31"

from . import experiments as experiments
from .configs import RootConfig as RootConfig
from .configs import Settings as Settings
from .utils.logging import debug as debug
from .utils.logging import error as error
from .utils.logging import info as info
from .utils.logging import info_bold as info_bold
from .utils.logging import warning as warning




import logging
import os
from logging import FileHandler, LogRecord, StreamHandler

from rich.console import Console
from rich.text import Text


class RichConsoleHandler(StreamHandler):  # type: ignore
    """
    Logging handler for rich console output.
    """

    def __init__(self) -> None:
        super().__init__()  # type: ignore

        self.console = Console()

    def emit(self, record: LogRecord) -> None:
        try:
            self.console.print(self.format(record))  # type: ignore
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


class RichFileHandler(FileHandler):
    """
    Logging handler for file output with stripped formatting. 
    """

    def __init__(self, filename: str) -> None:
        super().__init__(filename)

    def emit(self, record: LogRecord) -> None:
        record.msg = Text.from_markup(str(record.msg)).plain
        return super().emit(record)


LOG_LEVEL = logging.DEBUG if os.getenv('RUN_MODE', '').lower() == 'debug' else logging.INFO

log = logging.getLogger('main')
log.setLevel(LOG_LEVEL)


def debug(msg: str):
    log.debug(f':information_source: {msg}')


def info(msg: str):
    log.info(f'{msg}')


def info_bold(msg: str):
    log.info(f'[bold yellow]{msg}')


def warning(msg: str):
    log.warning(f':exclamation:[bold yellow]{msg}')


def error(msg: str):
    log.error(f':x: [bold red]{msg}')



import os
import random
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Type

import hydra
import numpy.random
import setproctitle
import torch
from dotenv import load_dotenv
from dotenv.main import find_dotenv
from omegaconf import OmegaConf

from .configs import RootConfig, Settings, register_configs
from .utils.logging import info, info_bold
from .utils.rundir import finish_rundir, setup_rundir

load_dotenv(find_dotenv(usecwd=True))


class Experiment(ABC):
    def __init__(self, config_path: str, settings_cls: Type[Settings], settings_group: Optional[str] = None) -> None:
        """
        Run an experiment from a provided entry point with minimal boilerplate code.

        Incorporates run directory setup and Hydra support.
        """

        self.settings_cls = settings_cls
        self.settings_group = settings_group
        self.root_cfg: RootConfig

        assert os.getenv('DATA_DIR') is not None, "Missing DATA_DIR environment variable."
        assert os.getenv('RESULTS_DIR') is not None, "Missing RESULTS_DIR environment variable."
        assert os.getenv('WANDB_PROJECT') is not None, "Missing WANDB_PROJECT environment variable."
        os.environ['DATA_DIR'] = str(Path(os.environ['DATA_DIR']).expanduser())
        os.environ['RESULTS_DIR'] = str(Path(os.environ['RESULTS_DIR']).expanduser())

        setup_rundir()
        self.before_entry()

        # Hydra will change workdir to the run dir before calling `self.main`
        register_configs(self.settings_cls, self.settings_group)
        hydra_decorator = hydra.main(config_path=config_path, config_name='root')
        hydra_decorator(self.entry)()

        self.finish()
        finish_rundir()

    @abstractmethod
    def before_entry(self) -> None:
        pass

    @abstractmethod
    def entry(self, root_cfg: RootConfig) -> None:
        """
        Experiment entrypoint. Called after initial setup with `cfg` populated by Hydra.

        Parameters
        ----------
        cfg : Config
            Top-level Hydra config for the experiment.
        """
        self.root_cfg = root_cfg
        self.cfg = root_cfg.cfg

        RUN_NAME = os.getenv('RUN_NAME')
        info_bold(f'\\[init] Run name --> {RUN_NAME}')
        info(f'\\[init] Loaded config:\n{OmegaConf.to_yaml(self.root_cfg, resolve=True)}')
        setproctitle.setproctitle(f'{RUN_NAME} ({os.getenv("WANDB_PROJECT")})')  # type: ignore

        self.seed_everything(root_cfg.cfg.seed)
        self.main()

    def seed_everything(self, seed: int) -> None:
        random.seed(seed)
        numpy.random.seed(seed)
        torch.manual_seed(seed)  # type: ignore
        torch.cuda.manual_seed_all(seed)

    @abstractmethod
    def main(self) -> None:
        pass

    @abstractmethod
    def finish(self) -> None:
        pass
